import os

import sys

# Dynamically get the directory of the current script
current_dir = os.path.dirname(os.path.abspath(__file__))

# Append the pipeline scripts directory to sys.path
sys.path.append(os.path.join(current_dir, 'pipeline', 'scripts'))

print("Current working directory:", os.getcwd())
print("Current directory:", current_dir)

import pandas as pd
import matplotlib

from pyairtable import Api
from dotenv import load_dotenv

from scripts.pipeline_utils import save_csv, get_table,airtable_to_df,response_to_df,run_report, main,retrieve_submissions_data

from google.analytics.data_v1beta import BetaAnalyticsDataClient
from google.analytics.data_v1beta.types import (
    DateRange,
    Dimension,
    Metric,
    RunReportRequest,
)

import os.path
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
import json

import datetime as dt
from datetime import datetime
from datetime import timedelta

load_dotenv()

class data_pipeline():
    def __init__(self,submission_directory,
                 save_directory, use_cached_data=False):
        self.submission_records = None
        self.use_cached_data = use_cached_data
        
        # Get the current working directory
        current_dir = os.getcwd()  # Gets the current working directory
        
        # Construct the submission directory as a variable
        self.submission_directory = submission_directory
        print(f'self.submission_directory:{self.submission_directory}')

        self.save_directory = save_directory
        print(f'self.save_directory:{self.save_directory}')

        self.twitter_data = None

    def set_use_cached_data(self, use_cached_data):
        self.use_cached_data = use_cached_data

    def save_data(self, data_struct, file):  
        # Define a custom serializer to handle Timestamp objects
        def custom_serializer(obj):
            if isinstance(obj, (pd.Timestamp, datetime)):
                return obj.isoformat()  # Convert Timestamp to string in ISO format
            raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")

        # Set the directory to 'pipeline/data/cleaned/'
        directory = self.save_directory
        
        # Ensure the directory exists
        if not os.path.exists(directory):
            os.makedirs(directory)  # Create the directory if it doesn't exist

        # Set the full file path
        path = os.path.join(directory, file)

        print(f'path: {path}')

        # Handle saving to JSON format
        if file.endswith('.json'):
            if isinstance(data_struct, dict):
                # Convert each DataFrame to a JSON-compatible format
                json_data_struct = {key: df.to_dict(orient='records') for key, df in data_struct.items()}
                with open(path, 'w') as f:
                    json.dump(json_data_struct, f, indent=4, default=custom_serializer)  # Use custom_serializer here
            else:
                raise ValueError("data_struct must be a dictionary for JSON saving.")

        # Handle saving to CSV format
        elif file.endswith('.csv'):
            if isinstance(data_struct, pd.DataFrame):
                data_struct.to_csv(path, index=False)
            else:
                raise ValueError("data_struct must be a DataFrame for CSV saving.")
            
    def load_data(self,file):
        directory = self.save_directory
        path = os.path.join(directory, file)

        # Load data from JSON file
        if file.endswith('.json'):
            with open(path, 'r') as f:
                json_data = json.load(f)
                # Convert the JSON data back to a dictionary of DataFrames
                data_struct = {key: pd.DataFrame(value) for key, value in json_data.items()}
                return data_struct

        # Load data from CSV file
        elif file.endswith('.csv'):
            # Load the CSV into a DataFrame
            data_frame = pd.read_csv(path)
            return data_frame  # Return the DataFrame directly

        else:
            raise ValueError("Unsupported file format. Please provide a .json or .csv file.")

    def get_airtable_data(self):

        if self.use_cached_data:
            return self.load_data('airtable_data.json')
        
        api = Api(os.environ['AIRTABLE_API_KEY'])

        submissions_records = get_table(api,os.environ['baseID'], os.environ['tableID'])
        submissions_records_df = airtable_to_df(submissions_records)

        contributor_records = get_table(api,os.environ['baseID'], os.environ['tableID'])
        contributor_records_df = airtable_to_df(contributor_records)

        project_records = get_table(api,os.environ['baseID'], os.environ['tableID'])
        project_records_df = airtable_to_df(project_records)

        data_struct = {
            "project_records": project_records_df,
            "contributor_records": contributor_records_df,
            "submissions_records": submissions_records_df
        }

        self.submission_records = submissions_records_df

        self.save_data(data_struct,'airtable_data.json')

        return data_struct
    
    def get_google_analytics_data(self):

        if self.use_cached_data:
            return self.load_data('google_analytics.json')

        GOOGLE_APPLICATION_CREDENTIALS = os.getenv('GOOGLE_APPLICATION_CREDENTIALS')
        property_id = os.getenv('property_id')

        twenty_eight_days = str(dt.date.today() - timedelta(28))
        seven_days = str(dt.date.today() - timedelta(7))
        twelve_months = str(dt.date.today() - timedelta(365))
        today = str(dt.date.today())

        cities = run_report(property_id,dimension="cityid",metric="activeUsers")
        users_timeseries = run_report(property_id,dimension="date",metric="activeUsers")

        new_v_return_7d = run_report(property_id,dimension='newVsReturning',metric='active7DayUsers',start_date=today)
        new_v_return_7d = new_v_return_7d[new_v_return_7d['newVsReturning'] != '']

        new_v_return_28 = run_report(property_id,dimension='newVsReturning',metric='active28DayUsers',start_date=today)
        new_v_return_28 = new_v_return_28[new_v_return_28['newVsReturning'] != '']

        devices_7d = run_report(property_id,dimension='platformDeviceCategory',metric='active7DayUsers',start_date=today)

        devices_28d = run_report(property_id,dimension='platformDeviceCategory',metric='active28DayUsers',start_date=today)

        devices_at = run_report(property_id,dimension='platformDeviceCategory',metric='activeUsers')

        source_7d = run_report(property_id,dimension='firstUserDefaultChannelGroup',metric='active7DayUsers',start_date=today)

        source_28d = run_report(property_id,dimension='firstUserDefaultChannelGroup',metric='active28DayUsers',start_date=today)

        source_at = run_report(property_id,dimension='firstUserDefaultChannelGroup',metric='activeUsers')

        views_7d = run_report(property_id,dimension='unifiedScreenClass',metric='screenPageViews', start_date=seven_days)

        views_28d = run_report(property_id,dimension='unifiedScreenClass',metric='screenPageViews', start_date=twenty_eight_days)

        views_at = run_report(property_id,dimension='unifiedScreenClass',metric='screenPageViews')

        data_struct = {
            "cities": cities,
            "users_timeseries": users_timeseries,
            "new_v_return_7d": new_v_return_7d,
            "new_v_return_28d": new_v_return_28,
            "devices_7d": devices_7d,
            "devices_28d": devices_28d,
            "devices_at":devices_at,
            "source_7d":source_7d,
            "source_28d":source_28d,
            "source_at":source_at,
            "views_7d":views_7d,
            "views_28d":views_28d,
            "views_at":views_at
        }

        self.save_data(data_struct,'google_analytics.json')

        # data_struct = self.load_data('google_analytics.json')

        return data_struct

    def get_sheets_data(self):

        SPREADSHEET_ID = os.getenv('SPREADSHEET_ID')
        SCOPES = [os.getenv('SCOPES')]
        token_path = os.getenv('token_path')
        creds_path = os.getenv('creds_path')

        print(f'token_path: {token_path}')

        if self.use_cached_data:
            return self.load_data('google_sheets.json')

        # Google Sheets API; includes Ghost NL Data
        log_df = main(SPREADSHEET_ID=SPREADSHEET_ID,
                      SCOPES=SCOPES,
                      token_path=token_path,
                      creds_path=creds_path,
                      SHEET_NAME="Issue Log Data",
                      skip_first_row=True)
        
        subscriber_df = main(SPREADSHEET_ID=SPREADSHEET_ID,
                      SCOPES=SCOPES,
                      token_path=token_path,
                      creds_path=creds_path,
                      SHEET_NAME='Subscriber CSV',
                      skip_first_row=False)
        
        issue_df = main(SPREADSHEET_ID=SPREADSHEET_ID,
                      SCOPES=SCOPES,
                      token_path=token_path,
                      creds_path=creds_path,
                      SHEET_NAME='Issue CSV',
                      skip_first_row=False)

        ON_twitter = main(SPREADSHEET_ID=SPREADSHEET_ID,
                      SCOPES=SCOPES,
                      token_path=token_path,
                      creds_path=creds_path,
                      SHEET_NAME='Twitter CSV',
                      skip_first_row=False)
        
        Spencer_twitter = main(SPREADSHEET_ID=SPREADSHEET_ID,
                      SCOPES=SCOPES,
                      token_path=token_path,
                      creds_path=creds_path,
                      SHEET_NAME='Spencer Twitter CSV',
                      skip_first_row=False)
        
        data_struct = {
            "log_df":log_df,
            "subscriber_df":subscriber_df,
            "issue_df":issue_df,
            "ON_twitter":ON_twitter,
            "Spencer_twitter":Spencer_twitter
        }

        twitter_data = {
            "ON_twitter":ON_twitter,
            "Spencer_twitter":Spencer_twitter
        }

        for key in data_struct.keys():
            print(data_struct[key])

        self.save_data(data_struct,'google_sheets.json')

        self.twitter_data = twitter_data

        return data_struct
    
    def get_twitter_data(self):
        # For now we get from local folder

        if self.twitter_data is None:
            sheets = self.load_data('google_sheets.json')
            ON_twitter = sheets['ON_twitter']
            Spencer_twitter = sheets['Spencer_twitter']

            twitter_data = {
            "ON_twitter":ON_twitter,
            "Spencer_twitter":Spencer_twitter
        }
            
            self.twitter_data = twitter_data
            
        
        return self.twitter_data
    
    def get_substack_data(self,path1,
                          path2):
        substack1 = pd.read_csv(path1)
        substack2 = pd.read_csv(path2)

        data_struct = {
            'substack1':substack1,
            'substack2':substack2
        }

        return data_struct
    
    def get_submissions(self, start_date, end_date):
        # Use the current directory as the base if submission_directory is not provided
        self.get_airtable_data()
        
        if self.use_cached_data:
            data = self.load_data('airtable_data.json')
            self.submission_records = data['submissions_records']

        submissions = retrieve_submissions_data(self.submission_records, start_date, end_date, self.submission_directory)
        return submissions



        












    


# E:\Projects\ournetwork\pipeline\setup.py
from setuptools import setup, find_packages

setup(
    name="modular_pipeline",
    version="0.1.3",
    description="A library for building modular pipelines in Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Brandyn Hamilton",
    author_email="brandynham1120@gmail.com",
    url="https://github.com/BrandynHamilton/Data-Pipeline",
    packages=find_packages(include=["pipeline", "pipeline.*"]),  # Automatically finds subpackages
    install_requires=[
        "pandas",
        "numpy",
        "plotly",
        "kaleido",
        "pyairtable",
        "python-dotenv",
        "google-analytics-data",
        "google-auth-oauthlib",
        "google-api-python-client"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    include_package_data=True,
    zip_safe=False,
)



import os
import pandas as pd
import matplotlib

import requests
from IPython.display import Image, display

from pyairtable import Api
from dotenv import load_dotenv

from google.analytics.data_v1beta import BetaAnalyticsDataClient
from google.analytics.data_v1beta.types import (
    DateRange,
    Dimension,
    Metric,
    RunReportRequest,
)

import os.path
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

import datetime as dt
from datetime import timedelta

def save_csv(df, name):
    path = f'../data/{name}.csv'
    df.to_csv(path,index=False)

def get_table(api, baseid, tableid):
    table = api.table(baseid,tableid)
    records = table.all()
    return records

def airtable_to_df(at,fields_col='fields',cols_to_keep=['id', 'createdTime']):
    at_copy = at.copy()
    df = pd.DataFrame(at_copy)
    fields_df = pd.json_normalize(df[fields_col])
    df = pd.concat([df[cols_to_keep], fields_df], axis=1)
    return df

def response_to_df(response):
    # Extract column names from the dimension and metric headers
    dimension_names = [dim.name for dim in response.dimension_headers]
    metric_names = [met.name for met in response.metric_headers]
    
    # Create lists to store the data
    rows_data = []
    
    # Loop through the rows in the response and collect values
    for row in response.rows:
        row_dict = {}
        for idx, dim_value in enumerate(row.dimension_values):
            row_dict[dimension_names[idx]] = dim_value.value
        for idx, met_value in enumerate(row.metric_values):
            row_dict[metric_names[idx]] = met_value.value
        rows_data.append(row_dict)
    
    # Create a DataFrame from the collected data
    df = pd.DataFrame(rows_data)

    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.sort_values(by='date', inplace=True)
    
    return df

def run_report(property_id,dimension=None,metric=None, start_date='2022-01-01',end_date='today'):
    """Runs a simple report on a Google Analytics 4 property."""
    # TODO(developer): Uncomment this variable and replace with your
    #  Google Analytics 4 property ID before running the sample.
    # property_id = "YOUR-GA4-PROPERTY-ID"

    # Using a default constructor instructs the client to use the credentials
    # specified in GOOGLE_APPLICATION_CREDENTIALS environment variable.
    client = BetaAnalyticsDataClient()

    request = RunReportRequest(
        property=f"properties/{property_id}",
        dimensions=[
            Dimension(name=dimension),
        ],
        metrics=[
            Metric(name=metric),
        ],
        date_ranges=[DateRange(start_date=start_date, end_date=end_date)],
)

    response = client.run_report(request)

    if len(response.rows) == 0:
        print("No data found in the API response.")
    else:
        print("Data found:", response.rows)

    df = response_to_df(response)

    return df

def retrieve_submissions_data(submissions_records_df, start_date, end_date, submission_directory):
    # Select relevant columns and drop rows with all NaN values in the CSV columns
    csvs = submissions_records_df[['id', 'Project','Created', 'Image #1', 'Image #2', 'Image #3', 
                                    'Image 1 CSV File', 'Image 2 CSV File', 'Image 3 CSV File']].dropna(
        subset=['Image 1 CSV File', 'Image 2 CSV File', 'Image 3 CSV File'], how='all')
    print(f'test')
    # Convert 'Created' to datetime and filter the DataFrame
    csvs['Created'] = pd.to_datetime(csvs['Created'])
    csvs.sort_values(by='Created', ascending=False, inplace=True)

    csvs_filtered = csvs[(csvs['Created'] >= start_date) & (csvs['Created'] <= end_date)]

    # Create a list to hold the structured data
    submissions_data = []

    for _, row in csvs_filtered.iterrows():
        # Prepare image and CSV data
        images = {}
        csv_files = {}

        for i in range(1, 4):
            # Access images
            image_key = f'Image #{i}'
            image_info = row[image_key]

            # Check if image_info is a non-empty list
            if isinstance(image_info, list) and image_info:
                image_url = image_info[0]['url']  # Extract URL
                image_filename = image_info[0]['filename']  # Extract filename
                images[f'image_{i}'] = image_url
                # Download the image
                download_file(image_url, submission_directory, image_filename)

        for i in range(1, 4):
            # Access CSV files
            csv_key = f'Image {i} CSV File'
            csv_info = row[csv_key]

            # Check if csv_info is a non-empty list
            if isinstance(csv_info, list) and csv_info:
                csv_url = csv_info[0]['url']  # Extract URL
                csv_filename = csv_info[0]['filename']  # Extract filename
                csv_files[f'csv_{i}'] = csv_filename  # Store the filename instead of the URL
                # Download the CSV file
                print(f'csv file: {csv_filename}')
                download_file(csv_url, submission_directory, csv_filename)

        # Constructing a dictionary for each submission
        submission_dict = {
            'dt': row['Created'],
            'project': row['Project'],
            'id': row['id'],
            'image': images,
            'data': csv_files  # This now contains the filenames of the CSV files
        }
        print(f'submission_dict: {submission_dict}')
        submissions_data.append(submission_dict)

    return submissions_data

def download_file(url, directory, filename):
    """Download a file from a URL and save it to the specified directory."""
    file_path = os.path.join(directory, filename)

    # Download the file
    response = requests.get(url)

    # Check if the request was successful
    if response.status_code == 200:
        # Write the content to a local file
        with open(file_path, 'wb') as f:
            f.write(response.content)
        print(f"Downloaded {filename} successfully.")
    else:
        print(f"Failed to download {filename}. Status code: {response.status_code}")

def main(SPREADSHEET_ID, SHEET_NAME,SCOPES,token_path,creds_path,skip_first_row=False):
    """Retrieves and prints the entire table from a specific sheet in Google Sheets."""
    creds = None

    token_path = os.path.join("config", f"{token_path}")

    creds_path = os.path.join("config", f"{creds_path}")


    print(f'token_path: {token_path}')
    print(f'scopes: {SCOPES}')
    print(f'spreadsheet:{SPREADSHEET_ID}')

    # if os.path.exists(token_path):
    #     creds = Credentials.from_authorized_user_file(token_path, SCOPES)
    # if not creds or not creds.valid:
    #     if creds and creds.expired and creds.refresh_token:
    #         creds.refresh(Request())
    #     else:
    #         flow = InstalledAppFlow.from_client_secrets_file(creds_path, SCOPES)
    #         creds = flow.run_local_server(port=0)
    #     with open(token_path, "w") as token:
    #         token.write(creds.to_json())

    if os.path.exists("token.json"):
        creds = Credentials.from_authorized_user_file("token.json", SCOPES)
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file("credentials.json", SCOPES)
            creds = flow.run_local_server(port=0)
        with open("token.json", "w") as token:
            token.write(creds.to_json())

    try:
        service = build("sheets", "v4", credentials=creds)

        result = service.spreadsheets().values().get(
            spreadsheetId=SPREADSHEET_ID, range=f"'{SHEET_NAME}'!A1:Z"
        ).execute()

        values = result.get("values", [])

        if not values:
            print("No data found.")
            return

        # Print the entire table from the specific sheet
        print("Retrieved table data from the specified sheet:")
        for row in values:
            print(row)

        # Convert the values to a DataFrame
        df = convert_to_dataframe(values,skip_first_row=skip_first_row)
        if df is not None:
            print("Converted DataFrame:")
            print(df.head())  # Display the first few rows of the DataFrame

            return df

    except HttpError as err:
        print(f"An error occurred: {err}")

def convert_to_dataframe(values, skip_first_row=False):
    if not values or len(values) < 3:  # Check for at least 3 rows (1 header + 1 data row)
        print("Not enough data to convert to DataFrame.")
        return None

    # Optionally skip the first row and use the second row as headers
    if skip_first_row:
        headers = values[1]  # Use the second row as headers
        data = values[2:]    # Remaining rows are the data
    else:
        headers = values[0]  # Use the first row as headers
        data = values[1:]    # Remaining rows are the data

    print(f'headers: {headers}')

    # Initialize an empty list to hold processed rows
    processed_data = []

    # Check for consistent row lengths and handle mismatches
    for row in data:
        if len(row) == len(headers):
            processed_data.append(row)
        else:
            # Create a new row that matches the number of headers, fill missing spots with None
            new_row = [None] * len(headers)
            for i in range(min(len(row), len(headers))):  # Only fill up to the number of headers
                new_row[i] = row[i]
            processed_data.append(new_row)

    # Create the DataFrame
    df = pd.DataFrame(processed_data, columns=headers)

    # Optionally: clean up the DataFrame (e.g., remove empty rows)
    df = df.dropna(how='all')  # Drop rows where all elements are NaN

    return df

def get_submission_df(submissions_data, submission_num, submission_directory):
    # Get the submission data
    try:
        submission = submissions_data[submission_num]
    except IndexError:
        raise ValueError("Submission number is out of range.")

    # Initialize dictionaries to store DataFrames and images
    project = submission['project']
    data_struct = {}
    images = {}

    # Loop through CSVs and corresponding images
    for i in range(1, 4):  # Assuming 'csv_1', 'csv_2', 'csv_3' and 'image_1', 'image_2', 'image_3'
        csv_key = f'csv_{i}'
        img_key = f'image_{i}'

        # Load CSV or Excel based on the file extension
        csv_filename = submission['data'].get(csv_key)
        if csv_filename:
            csv_path = os.path.join(submission_directory, csv_filename)
            try:
                # Check if the file is an Excel file
                if csv_filename.endswith('.xlsx') or csv_filename.endswith('.xls'):
                    df = pd.read_excel(csv_path)
                else:
                    df = pd.read_csv(csv_path)
                data_struct[csv_filename] = df
            except FileNotFoundError:
                print(f"Warning: File {csv_path} not found.")
            except pd.errors.EmptyDataError:
                print(f"Warning: File {csv_path} is empty.")
            except UnicodeDecodeError:
                print(f"Error: {csv_path} contains invalid encoding.")
            except Exception as e:
                print(f"Error loading {csv_path}: {e}")

        # Get image URL
        img_url = submission['image'].get(img_key)
        if img_url:
            images[img_key] = img_url

    file_names = list(data_struct.keys())
    image_values = list(images.values())

    submission = {
        "project": project,
        "file_names": file_names,
        "image_urls": image_values
    }

    return submission


def show_file_and_img(submission,index=0):
    # Show the file and image
    for key in submission.keys():
        if key == 'project':
            print(f"Project: {submission[key]}")
        if key == 'file_names':
            print(f"File Name: {submission[key][index]}")
        elif key == 'image_urls':
            display(Image(url=submission[key][index]))

def get_files(submission):
    keys_list = list(submission['file_names'])

    # Accessing the keys
    first_key = keys_list[0] if len(keys_list) > 0 else None
    second_key = keys_list[1] if len(keys_list) > 1 else None
    third_key = keys_list[2] if len(keys_list) > 2 else None

    files = {
        'first_file': first_key if first_key else None,
        'second_file': second_key if second_key else None,
        'third_file': third_key if third_key else None,
    }

    return files